#ABC = Abstract Base Class
from abc import ABC, abstractmethod

class StorageABC(ABC):
    @abstractmethod
    def __init__(self, db_name: str):
        self.db_name = db_name

    #Methods for key-value storage
    @abstractmethod
    def set(self, table_name: str, key: str, value: any):
        pass

    @abstractmethod
    def set_all(self, table_name:str, full_data: dict):
        pass

    @abstractmethod
    def remove(self, table_name: str, key: str):
        pass

    @abstractmethod
    def update(self, table_name:str, update_data: dict):
        pass

    @abstractmethod
    def get(self, table_name: str, key: str):
        pass

    @abstractmethod
    def get_all(self, table_name: str, keys: (list, None)=None):
        pass


import logging
import uuid
from .utils.exec_tree import create_empty_data_slot
from .utils.logger import get_default_logger
from .utils.find import find_by_attr
from .lib.BreakingHub import BreakingHub
from .lib.Store import Store

workflow_default_logger = get_default_logger('Workflow', logging.INFO)

class MainExecutor:
    def __init__(self, settings={}):
        self.is_running = True
        self.running_status = 'idle'
        self.settings = settings
        self.max_execution_limit = (settings or {}).get('max_execution_limit') or 10
        self.breaking_hub = BreakingHub(
            breaking_handler = self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )
        self.store = Store()
        self.logger = settings.get('logger', workflow_default_logger)
        # 已注册的执行器类型
        self.registed_executors = {}
        self.chunks_map = {}

    def start(self, runtime_data: dict):
        entries = runtime_data.get('entries') or []
        self.chunks_map = runtime_data.get('chunk_map') or {}
        self._reset_temp_status()
        self.running_status = 'start'

        self._execute_main(entries)
        self.running_status = 'end'

    def regist_executor(self, name: str, executor):
        """
        注册执行器，传入执行器的名称及executor
        """
        self.registed_executors[name] = executor

    def unregist_executor(self, name: str):
        """
        取消注册执行器，传入执行器的名称
        """
        if name in self.registed_executors:
            del self.registed_executors[name]

        return self

    def handle_command(self, data):
        """
        用于接收处理外部指令
        """
        if data is None or self.is_running == False or data["dataset"] is None:
            return

        command = data["dataset"]["command"]
        command_data = data["dataset"]["data"]
        if command == "input":
            if command_data and command_data["content"] and command_data['schema']:
                self.input(command_data["content"], command_data['schema'])
        elif command == "destroy":
            self.is_running = False
    
    def _execute_main(self, entries: list):
        """执行入口"""
        for entry_chunk in entries:
            slow_tasks = []
            self._execute_partial(
                entry_chunk,
                slow_tasks=slow_tasks,
                executing_ids=[],
                visited_record=[]
            )
            # 如果执行完成后，还有待执行任务，则需手动执行
            if len(slow_tasks) > 0:
                self._execute_slow_tasks(slow_tasks)
        

    def _execute_partial(
            self,
            chunk,
            slow_tasks = [], # 缓执行任务
            executing_ids = [], # 执行中的任务
            visited_record = [], # 已执行记录
        ):
        """执行某个分组，在单次分组里，不会存在循环节点，遇到循环节点，会另起一个分组执行"""

        # 针对循环，存入缓执行组中，延后执行
        if chunk['id'] in visited_record:
            if chunk.get('loop_entry') == True:
                slow_tasks.append(chunk)
            return

        # 1、校验上游依赖是否就绪
        if not self._check_dep_ready(chunk):
            self.logger.info(f"'{self._get_chunk_title(chunk)}' dependencies not ready.")
            return


        # 2、执行当前 chunk
        execute_id = uuid.uuid4()
        executing_ids.append(execute_id)
        self.logger.info(f"Execute '{self._get_chunk_title(chunk)}'")
        exec_value =  self._exec_chunk_core(chunk)
        self.breaking_hub.recoder(chunk)  # 更新中断器信息
        visited_record.append(chunk['id']) # 更新执行记录
        
        # 3、执行完成后，继续下游任务
        # 提取当前执行的结果，尝试将当前执行结果注入到下游的运行依赖插槽上 next_chunk['deps'][]['data_slot'] = 'xxx'
        for next_info in chunk['next_chunks']:
            next_chunk = self.chunks_map.get(next_info['id'])
            if not next_chunk:
                continue

            # 针对目标 chunk 的目标 handle 的插槽位，注入最新的值（注意如果是条件连接线，需要在条件满足时才更新）
            for next_rel_handle in next_info['handles']:
                source_handle = next_rel_handle['source_handle']
                target_handle = next_rel_handle['handle']
                source_value = exec_value.get(source_handle) if isinstance(exec_value, dict) else exec_value

                # 有条件的情况下，仅在条件满足时，才更新下游节点的数据
                condition_call = next_rel_handle.get('condition')
                if condition_call:
                    judge_res = condition_call(source_value)
                    if judge_res != True:
                        continue
                
                next_chunk_target_dep = find_by_attr(next_chunk['deps'], 'handle', target_handle)
                if next_chunk_target_dep:
                    next_chunk_target_dep['data_slot'] = {
                        'is_ready': True,
                        'updator': chunk['id'],
                        'value': source_value,
                    }

        # 依次尝试执行下游节点
        for next_info in chunk['next_chunks']:
            next_chunk = self.chunks_map.get(next_info['id'])
            if not next_chunk:
                continue

            self._execute_partial(next_chunk, slow_tasks, executing_ids, visited_record)
        
        # 任务执行完后，清理执行中的状态
        executing_ids.remove(execute_id)
        # 尝试执行缓执行任务
        if len(executing_ids) == 0:
            self._execute_slow_tasks(slow_tasks)
            slow_tasks.clear()

    def _execute_slow_tasks(self, slow_tasks):
        """尝试执行低优任务"""
        if len(slow_tasks) == 0:
            return

        for chunk in slow_tasks:
            # 先清空下游所有节点数据
            self._chunks_clean_walker(chunk)
            # 再启动执行
            self._execute_partial(
                chunk,
                slow_tasks=[],
                executing_ids=[],
                visited_record=[]
            )
            # if len(other_slow_tasks) > 0:
            #     self._execute_slow_tasks(other_slow_tasks)

    def _get_chunk_executor(self, name: str):
        """
        根据类型名称获取执行器
        """
        return self.registed_executors.get(name)

    def _reset_temp_status(self):
        """
        重置临时状态
        """
        self.running_status = 'idle'
        # 中断处理
        self.breaking_hub = BreakingHub(
            breaking_handler=self._handle_breaking,
            max_execution_limit=self.max_execution_limit
        )

    def _get_chunk_res_setter_key(self, chunk):
        """
        获取存储数据的 key（需要考虑循环调用的场景）
        """
        return f"{chunk['id']}_{self.breaking_hub.get_counts(chunk) + 1}"

    def _get_chunk_res_getter_key(self, chunk):
        """
        获取读取数据的 key（需要考虑循环调用的场景）
        """
        return f"{chunk['id']}_{self.breaking_hub.get_counts(chunk)}"

    def _exec_chunk_core(self, chunk):
        """
        执行任务（执行到此处的都是上游数据已就绪了的）
        """
        # 简化参数
        deps_dict = {}
        for dep in chunk['deps']:
            deps_dict[dep['handle']] = dep['data_slot']['value']

        # 交给执行器执行
        executor_type = chunk['data']['type']
        chunk_executor = self._get_chunk_executor(executor_type) or chunk.get('executor')
        exec_res = None
        try:
            exec_res = chunk_executor(deps_dict, self.store)
        except Exception as e:
            self.logger.error(f"Node Execution Exception: '{self._get_chunk_title(chunk)}'({chunk['id']}) {e}")
            # 主动中断执行
            raise Exception(e)

        return exec_res

    def _check_dep_ready(self, chunk):
        """监测依赖是否就绪"""
        deps = chunk.get('deps', [])
        if len(deps) == 0:
            return True

        for dep in deps:
            if dep['data_slot']['is_ready'] == False:
                return False
        return True

    def _chunks_clean_walker(self, root_chunk):
        """尝试对某个节点以下的分支做一轮清理工作"""

        visited_record = []
        def clean_core(chunk):
            for next_info in chunk['next_chunks']:
                next_chunk = self.chunks_map.get(next_info['id'])
                if not next_chunk:
                    continue

                if (next_chunk['id'] in visited_record) or (next_chunk['id'] == root_chunk['id']):
                    continue

                visited_record.append(next_chunk['id'])
                effect_handles = [handle_desc['handle'] for handle_desc in next_info['handles']]
                for dep in next_chunk['deps']:
                    data_slot = dep['data_slot']
                    # 如果当前的输入句柄受上游影响了，则重置
                    if (dep['handle'] in effect_handles) and (data_slot['updator'] == '' or data_slot['updator'] == chunk['id']):
                        dep['data_slot'] = create_empty_data_slot(next_chunk)

                clean_core(next_chunk)

        clean_core(root_chunk)
        
    
    def _get_chunk_title(self, chunk):
        return chunk["data"]["title"] or f'chunk-{chunk["data"]["id"]}' or 'Unknow chunk'

    def _handle_breaking(self, chunk, type):
        """处理中断"""
        self.logger.error(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")
        # 中断之前处理相关逻辑
        # 主动中断执行
        raise Exception(
            f"Exceeded maximum execution limit: {self._get_chunk_title(chunk)}")



import re
import json
import json5
from .transform import find_json, to_json_desc

def check_structure(origin: any, compare_target: any, position: str=""):
    errors = []
    if isinstance(origin, dict):
        for key in origin.keys():
            current_position = position + "." + key
            if key not in compare_target:
                errors.append(current_position[1:])
            check_result = check_structure(origin[key], compare_target[key], current_position)
            if check_result != True:
                errors.extend(check_result)
    if isinstance(origin, list):
        for index, item in enumerate(origin):
            current_position = position + "." + str(index)
            if index in enumerate(compare_target):
                check_result = check_structure(item, compare_target[index], current_position)
                if check_result != True:
                    errors.extend(check_result)
    if len(errors) == 0:
        return True
    else:
        return errors

async def fix_json(origin: str, input_dict: any, output_dict: any, request: object, *, is_debug: bool=False, errors = list):
    try:
        fixed_result = await request\
            .input({
                "error JSON String": origin ,
                "error info": errors,
            })\
            .output('FIXED {error JSON String} JSON STRING ONLY WITHOUT EXPLANATION that can be parsed by Python')\
            .start_async()
        json_string = find_json(fixed_result)
        if is_debug:
            print("[Cleaned JSON String]:\n", json_string)
            print("\n--------------------------\n")
        fixed_result = json5.loads(json_string)
        if is_debug:
            print("[Parse JSON to Dict] Done")
            print("\n--------------------------\n")
        return fixed_result
    except Exception as e:
        raise Exception(f"[Agent Request] Error still occured when try to fix JSON decode error: { str(e) }\nOrigin JSON String:\n{ json_string }")

async def load_json(origin: str, input_dict: any, output_dict: any, request: object, *, is_debug: bool=False):
    try:
        json_string = find_json(origin)
        if is_debug:
            print("[Cleaned JSON String]:\n", json_string)
            print("\n--------------------------\n")
        parsed_dict = json5.loads(json_string)
        return parsed_dict
    except Exception as e:
        if is_debug:
            print("[JSON Decode Error Occurred] Start Fixing Process...")
        return await fix_json(origin, input_dict, output_dict, request, is_debug = is_debug, errors = [str(e)])

def find_and_load_json(origin: str):
    try:
        json_string = find_json(origin)
        return json5.loads(json_string)
    except Exception as e:
        return 'Error'


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2023/11/10 15:55
# @Author  : yongjie.su
# @File    : MiniMax.py
# @Software: PyCharm
import json
import aiohttp
from .utils import RequestABC, to_instruction, to_json_desc, to_prompt_structure
from Agently.utils import RuntimeCtxNamespace

minimax_chat_pro = {
    "default_url": "https://api.minimax.chat/v1/text/chatcompletion_pro",
    "default_model_name": "abab5.5-chat",
    "default_max_token": 16384
}

minimax_chat = {
    "default_url": "https://api.minimax.chat/v1/text/chatcompletion",
    "default_model_name": "abab5-chat",
    "default_max_token": 6144
}

minimax_t2a = {
    "default_url": "https://api.minimax.chat/v1/text_to_speech",
    "default_model_name": "speech-01"
}

minimax_t2a_pro = {
    "default_url": "https://api.minimax.chat/v1/t2a_pro",
    "default_model_name": "speech-01"
}

minimax_embeddings = {
    "default_url": "https://api.minimax.chat/v1/embeddings",
    "default_model_name": "embo-01",
    "default_max_token": 4096
}


class MiniMax(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = 'MiniMax'
        _request_type = self.request.request_runtime_ctx.get("request_type", "chat")
        self.request_type = _request_type if _request_type is not None else "chat"
        self.model_settings = RuntimeCtxNamespace(f"model.{self.model_name}", self.request.settings)

    def construct_request_messages(self):
        # init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get(
            "prompt.general_instruction")
        if general_instruction_data:
            request_messages.append({"role": "user",
                                     "content": f"[重要指导说明]\n{to_instruction(general_instruction_data)}"})
            request_messages.append({"role": "assistant", "content": "OK"})
        # - role
        role_data = self.request.request_runtime_ctx.get("prompt.role")
        if role_data and self.request_type == "chat":
            request_messages.append(
                {"role": "user", "content": f"[角色及行为设定]\n{to_instruction(role_data)}"})
            request_messages.append({"role": "assistant", "content": "OK"})
        # - user info
        user_info_data = self.request.request_runtime_ctx.get("prompt.user_info")
        if user_info_data and self.request_type == "chat":
            request_messages.append(
                {"role": "user", "content": f"[用户信息]\n{to_instruction(user_info_data)}"})
            request_messages.append({"role": "assistant", "content": "OK"})
        # - headline
        headline_data = self.request.request_runtime_ctx.get("prompt.headline")
        if headline_data:
            request_messages.append(
                {"role": "user", "content": f"[主题及摘要]{to_instruction(headline_data)}"})
            request_messages.append({"role": "assistant", "content": "OK"})
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get("prompt.chat_history")
        if chat_history_data:
            request_messages.extend(chat_history_data)
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get("prompt.input")
        prompt_information_data = self.request.request_runtime_ctx.get("prompt.information")
        prompt_instruction_data = self.request.request_runtime_ctx.get("prompt.instruction")
        prompt_output_data = self.request.request_runtime_ctx.get("prompt.output")
        # --- only input
        if not prompt_input_data and not prompt_information_data and not prompt_instruction_data and not prompt_output_data:
            raise Exception(
                "[Request] Missing 'prompt.input', 'prompt.information', 'prompt.instruction', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        if prompt_input_data and not prompt_information_data and not prompt_instruction_data and not prompt_output_data:
            request_messages.append({"role": "user", "content": to_instruction(prompt_input_data)})
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[输入]"] = to_instruction(prompt_input_data)
            if prompt_information_data:
                prompt_dict["[补充信息]"] = to_instruction(prompt_information_data)
            if prompt_instruction_data:
                prompt_dict["[处理规则]"] = to_instruction(prompt_instruction_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[输出要求]"] = {
                        "TYPE": "JSON can be parsed in Python",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[输出要求]"] = str(prompt_output_data)
            request_messages.append(
                {"role": "user", "content": to_prompt_structure(prompt_dict, end="[输出]:\n")})
        return request_messages

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options")
        if not isinstance(options, dict):
            raise ValueError("The value type of 'model_settings.options' must be a dict.")
        auth = self.model_settings.get_trace_back("auth")
        if not isinstance(auth, dict):
            raise ValueError("The value type of 'model_settings.auth' must be a dict.")
        api_key = auth.get('api_key')
        group_id = auth.get('group_id')
        if not api_key:
            raise ValueError("The value of 'api_key' is invalid.")
        body = {
            "stream": self.request.request_runtime_ctx.get("stream", False)
        }
        prompt = self.request.request_runtime_ctx.get("prompt")
        text = self.request.request_runtime_ctx.get("prompt.input")
        if self.request_type.lower() == "chat_pro":
            base_url = options.get('base_url', minimax_chat_pro.get('default_url'))
            body.update({
                "model": options.get('model', minimax_chat_pro.get('default_model_name')),
                "bot_setting": [
                    {
                        "bot_name": "MM智能助理",
                        "content": "MM智能助理是一款由MiniMax自研的，没有调用其他产品的接口的大型语言模型。MiniMax是一家中国科技公司，一直致力于进行大模型相关的研究。",
                    }
                ],
                "messages": [
                    {
                        "sender_type": "USER",
                        "sender_name": "小明",
                        "text": "帮我用英文翻译下面这句话：我是谁"
                    }
                ],
                "reply_constraints": {
                    "sender_type": "BOT",
                    "sender_name": "MM智能助理"
                },

                "tokens_to_generate": 1034,
                "temperature": 0.01,
                "top_p": 0.95,
            })
        elif self.request_type.lower() == "chat":
            base_url = options.get('base_url', minimax_chat.get('default_url'))
            body.update({
                "model": options.get('model', minimax_chat.get('default_model_name')),
                "prompt": "你是一个专家",
                "role_meta": {
                    "user_name": "USER",
                    "bot_name": "BOT"
                },
                "messages": [{
                    "sender_type": self.request.request_runtime_ctx.get("sender_type", "USER"),
                    "text": text,

                }],
                "temperature": options.get('temperature', 0.5)
            })
        elif self.request_type.lower() == "embeddings":
            base_url = options.get('base_url', minimax_embeddings.get('default_url'))
            body.update({
                "model": options.get('model', minimax_embeddings.get('default_model_name')),
                "texts": text,
                "type": options.get('type', 'query')
            })

        else:
            raise Exception(f"This type '{self.request_type}' is currently not supported.")
        # 统一构造请求数据结构
        request_data = {
            "url": base_url,
            "params": {"GroupId": group_id},
            "headers": {
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
            },
            "body": body
        }
        return request_data

    async def request_model(self, request_data: dict, timeout=60):
        url = request_data.get('url')
        headers = request_data.get('headers')
        body = request_data.get('body')
        params = request_data.get('params')
        proxy = request_data.get("proxy") if "proxy" in request_data else None
        async with aiohttp.ClientSession() as session:
            async with session.post(url, params=params, json=body, headers=headers, proxy=proxy,
                                    timeout=timeout) as response:
                response = await response.text()
                response = json.loads(response)
                if response["base_resp"]["status_code"] != 0:
                    raise Exception(response["base_resp"])
                return response.get('choices', [])

    def broadcast_response(self, response_generator):
        response_message = {"role": "assistant", "content": ""}
        full_response_message = {}
        for part in response_generator:
            full_response_message = dict(part)
            delta = part["text"]
            response_message["content"] += delta
            yield ({"event": "response:delta_origin", "data": part})
            yield ({"event": "response:delta", "data": delta})
        full_response_message["result"] = response_message["content"]
        yield ({"event": "response:done_origin", "data": full_response_message})
        yield ({"event": "response:done", "data": response_message["content"]})

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }


def export():
    return "MiniMax", MiniMax



from .StartExecutor import start_executor
from ..MainExecutor import MainExecutor

def mount_built_in_executors(main_executor: MainExecutor):
    """
    挂载内置的执行器
    """
    main_executor.regist_executor('Start', start_executor)


from .DataOps import DataOps

class StorageEditor(DataOps):
    def __init__(self, target_data: (dict, None), *, storage: object, table_name: str):
        self.storage = storage
        self.table_name = table_name
        super().__init__(target_data = target_data)

    def save(self):
        final_data = self.get()
        return self.storage.set_all(self.table_name, final_data)

class StorageDelegate(object):
    def __init__(self, *, db_name: str = "default", plugin_manager: object, settings: object):
        self.db_name = db_name
        self.plugin_manager = plugin_manager
        self.settings = settings
        return

    def set_storage_type(self, storage_type: str):
        self.settings.set("storage_type", storage_type)
        return self

    def __get_storage_plugin(self):
        storage_type = self.settings.get_trace_back("storage_type")
        return self.plugin_manager.get("storage", storage_type)(db_name = self.db_name)

    def set(self, table_name: str, key: str, value: any):
        self.__get_storage_plugin().set(table_name, key, value)
        return self

    def set_all(self, table_name: str, full_data: dict):
        self.__get_storage_plugin().set_all(table_name, full_data)
        return self

    def remove(self, table_name: str, key: str):
        self.__get_storage_plugin().remove(table_name, key)
        return self

    def update(self, table_name:str, update_data: dict):
        self.__get_storage_plugin().update(table_name, update_data)
        return self

    def get(self, table_name: str, key: str):
        return self.__get_storage_plugin().get(table_name, key)

    def get_all(self, table_name: str, keys: (list, None)=None):
        return self.__get_storage_plugin().get_all(table_name, keys)

    def table(self, table_name: str):
        table_data = self.get_all(table_name)
        return StorageEditor(table_data, storage = self, table_name = table_name)


from ..Schema import Schema
from ..lib.constants import EXECUTOR_TYPE_NORMAL, EXECUTOR_TYPE_START

def prepare_data(schema: Schema):
    edges_target_map = {}
    edges_source_map = {}
    edges = schema.edges

    for edge in edges:
        edges_target_map.setdefault(edge['source'], []).append(edge)
        edges_source_map.setdefault(edge['target'], []).append(edge)

    return {
        'edges_source_map': edges_source_map,
        'edges_target_map': edges_target_map
    }

def resolve_runtime_data(schema: Schema):
    prepared_data = prepare_data(schema)

    def create_exec_chunk(chunk):
        next_chunks = []
        next_chunk_map = {}
        for target in prepared_data['edges_target_map'].get(chunk['id'], []):
            target_id = target['target']
            next_chunk = next_chunk_map.get(target_id)
            # 如之前没存入，则新建
            if not next_chunk:
                next_chunk = {
                    'id': target_id,
                    'handles': []
                }
                next_chunks.append(next_chunk)
                next_chunk_map[next_chunk['id']] = next_chunk
            # 往 chunk 中注入当前的 handle
            next_chunk['handles'].append({
                'handle': target['target_handle'],
                'source_handle': target['source_handle'],
                'condition': target.get('condition') or None
            })
                
        hanldes_desc = chunk.get('handles') or {}
        inputs_desc = hanldes_desc.get('inputs') or []
        return {
            'id': chunk['id'],
            'loop_entry': False, # 是否是循环的起点
            'next_chunks': next_chunks,
            'type': chunk.get('type') or EXECUTOR_TYPE_NORMAL,
            'executor': chunk.get('executor'),
            'deps': [
                {
                    'handle': input_desc.get('handle'),
                    # 运行时的依赖值，会随运行时实时更新，初始尝试从定义中取默认值
                    'data_slot': {
                        'is_ready': (input_desc.get('default') != None) or (chunk.get('type') == EXECUTOR_TYPE_START),
                        'updator': '', # 更新者
                        'value': input_desc.get('default')
                    }
                }
                for input_desc in inputs_desc
            ],
            'data': chunk
        }
    
    runtime_chunks_map = {}
    for chunk in schema.chunks:
        runtime_chunks_map[chunk['id']] = create_exec_chunk(chunk)
    
    # 标识循环起点
    def update_loop_walker(chunk, paths = []):
        for next_info in chunk.get('next_chunks'):
            next_chunk = runtime_chunks_map.get(next_info['id'])
            if next_chunk:
                if next_chunk['id'] in paths:
                    next_chunk['loop_entry'] = True
                    print(next_chunk['data']['title'])
                    return
                update_loop_walker(next_chunk, paths + [next_chunk['id']])

    entries = []
    for chunk in schema.chunks:
        if not prepared_data['edges_source_map'].get(chunk['id'], []):
            entry_chunk = runtime_chunks_map.get(chunk['id'])
            if entry_chunk:
                entries.append(entry_chunk)
                update_loop_walker(entry_chunk, [entry_chunk['id']])

    return {
        'entries': entries,
        'chunk_map': runtime_chunks_map
    }

def create_empty_data_slot(chunk):
    return {
        'is_ready': chunk.get('type') == EXECUTOR_TYPE_START,
        'updator': '',
        'value': None
    }


import json
from .RuntimeCtx import RuntimeCtx
from .transform import find_json

class PluginManager(object):
    def __init__(self, *, parent: object=None):
        self.plugins_runtime_ctx = RuntimeCtx(parent = parent.plugins_runtime_ctx if parent else None)

    def register(self, module_name: str, plugin_name: str, plugin: callable):
        if module_name == "$":
            raise Exception("[Plugin Manager] Module name can not be '$'.")
        self.plugins_runtime_ctx.set(f"{ module_name }.{ plugin_name }", plugin)
        return self
    
    def set_settings(self, keys_with_dots: str, value: any):
        self.plugins_runtime_ctx.set(f"$.settings.{keys_with_dots  }", value)
        return self

    def update_settings(self, settings: dict):
        for key, value in settings.items():
            if value.lower() == "true":
                settings[key] = True
            if value.lower() == "false":
                settings[key] = False
            json_string = find_json(value)
            if json_string != None and json_string != '':
                settings[key] = json.loads(json_string)
            self.set_settings(key, settings[key])
        return self

    def get_settings(self, keys_with_dots: str=None, default: str=None):
        if keys_with_dots != None:
            return self.plugins_runtime_ctx.get_trace_back(f"$.settings.{ keys_with_dots }", default)
        else:
            return self.plugins_runtime_ctx.get_trace_back("$.settings", default)

    def get(self, module_name: str, plugin_name: str=None):
        plugins = self.plugins_runtime_ctx.get_trace_back()
        if plugins == None:
            plugins = {}
        if module_name not in plugins:
            raise Exception(f"[Plugin Manager] Module '{ module_name }' is not in plugins runtime_ctx.")
        if plugin_name != None:
            if plugin_name not in plugins[module_name]:
                raise Exception(f"[Plugin Manager] Plugin '{ plugin_name }' is not in the plugins runtime_ctx of module '{ module_name }'.")
            return plugins[module_name][plugin_name]
        else:
            return plugins[module_name]

    def get_agent_component_list(self):
        agent_components = self.get("agent_component")
        return agent_components.keys()


import json
from datetime import datetime
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace, to_json_desc

class Tool(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.is_debug = lambda: self.agent.settings.get_trace_back("is_debug")
        self.settings = RuntimeCtxNamespace("plugin_settings.agent_component.Tool", self.agent.settings)
        self.tool_manager = self.agent.tool_manager
        self.tool_dict = {}
        self.must_call_tool_info = None
        self.call_tool_func = self.tool_manager.call_tool_func
        self.set_tool_proxy = self.tool_manager.set_tool_proxy

    def register_tool(self, tool_name: str, desc: any, args: dict, func: callable, *, categories: (str, list) = None, require_proxy: bool=False):
        self.tool_manager.register(tool_name, desc, args, func, categories=categories, require_proxy=require_proxy)
        self.tool_dict.update({ tool_name: { "tool_name": tool_name, "desc": desc, "args": args } })
        return self.agent

    def stop_tools(self, tool_name_list: (str, list)):
        if isinstance(tool_name_list, str):
            tool_name_list = [tool_name_list]
        for tool_name in tool_name_list:
            if tool_name in self.tool_dict:
                del self.tool_dict[tool_name]
        return self.agent

    def add_public_tools(self, tool_name_list: (str, list)):
        if isinstance(tool_name_list, str):
            tool_name_list = [tool_name_list]
        for tool_name in tool_name_list:
            tool_info = self.tool_manager.get_tool_info(tool_name, with_args=True)
            if tool_info:
                self.tool_dict.update({ tool_name: tool_info })
        return self.agent

    def add_public_categories(self, tool_category_list: (str, list)):
        if isinstance(tool_category_list, str):
            tool_category_list = [tool_category_list]
        tool_dict = self.tool_manager.get_tool_dict(categories=tool_category_list, with_args=True)
        for key, value in tool_dict.items():
            self.tool_dict.update({ key: value })
        return self.agent

    def add_all_public_tools(self):
        all_tool_dict = self.tool_manager.get_tool_dict(with_args=True)
        for key, value in all_tool_dict.items():
            self.tool_dict.update({ key: value })
        return self.agent

    def must_call(self, tool_name: str):
        tool_info = self.tool_manager.get_tool_info(tool_name, full=True)
        if tool_info:
            self.must_call_tool_info = tool_info
        else:
            raise Exception(f"[Agent Component] Tool-must call: can not find tool named '{ tool_name }'.")

    async def call_plan_func(self, tool):
        if self.is_debug():
            print("[Agent Component] Using Tools: Start tool using judgement...")
        tool_list = []
        for tool_name, tool_info in self.tool_dict.items():
            tool_list.append(tool_info)
        result = (
            await self.agent.worker_request
                .input({
                    "target": self.agent.request.request_runtime_ctx.get("prompt")
                })
                .info("current date", datetime.now().date().strftime("%Y-%B-%d"))
                .info("tools", to_json_desc(tool_list))
                .instruct("what tools to use for achieving {input.target}.\n * if use search tool, choose ONLY ONE SEARCH TOOL THAT FIT MOST\n * OUTPUT LANGUAGE SAME AS {input.target} IS USING.")
                .output({
                    "tools_using": [{
                        "purpose": ("String", "what question you want to use tool to solve?"),
                        "using_tool": (
                            {
                                "tool_name": ("String", "{tool_name} from {tools}"),
                                "args": ("according {args} requirement in {tools}", ),
                            },
                            "output null if do not need to use tool"
                        ),
                    }],
                })
                .start_async()
        )
        tool_results = {}
        for step in result["tools_using"]:
            if "using_tool" in step and isinstance(step["using_tool"], dict) and "tool_name" in step["using_tool"]:
                if self.is_debug():
                    print("[Using Tool]: ", step["using_tool"])
                tool_info = self.tool_manager.get_tool_info(step["using_tool"]["tool_name"], full=True)
                if tool_info:
                    tool_kwrags = step["using_tool"]["args"] if "args" in step["using_tool"] and isinstance(step["using_tool"]["args"], dict) else {}
                    if tool_info["require_proxy"]:
                        proxy = self.agent.settings.get_trace_back("proxy")
                        if proxy == None:
                            proxy = self.agent.tool_manager.get_tool_proxy()
                        if proxy:
                            tool_kwrags.update({ "proxy": proxy })
                    call_result = None
                    try:
                        call_result = self.call_tool_func(
                            tool_info["tool_name"],
                            **tool_kwrags
                        )
                    except Exception as e:
                        if self.is_debug():
                            print("[Tool Error]: ", e)
                    if call_result:
                        info_key = str(step["purpose"])
                        info_value = call_result["for_agent"] if isinstance(call_result, dict) and "for_agent" in call_result else call_result
                        tool_results[info_key] = info_value
                        if self.is_debug():
                            print("[Result]: ", info_key, info_value)
                else:
                    if self.is_debug():
                        print(f"[Result]: Can not find tool '{ step['using_tool']['tool_name'] }'")
        if len(tool_results.keys()) > 0:
            return tool_results
        else:
            return None

    def customize_call_plan(self, call_plan_func: callable):
        self.call_plan_func = call_plan_func
        return self.agent

    async def _prefix(self):
        if self.must_call_tool_info:
            if self.is_debug():
                print(f"[Agent Component] Using Tools: Must call '{ self.must_call_tool_info['tool_name'] }'")
            self.agent.request.request_runtime_ctx.remove("prompt.instruct")
            self.agent.request.request_runtime_ctx.remove("prompt.output")
            return {
                "information": {
                    "function_must_call": {
                        "tool_name": self.must_call_tool_info["tool_name"],
                        "desc": self.must_call_tool_info["desc"],
                        "args": self.must_call_tool_info["args"],
                    },
                },
                "output": {
                    "can_call": ("Boolean", "Is all information especially from {input} above enough to generate all arguments in {function_must_call.args}?"),
                    "args": ("if {can_call} == true, generate args dict according {function_must_call.args} requirement, else leave argument value as null"),                    
                    "question": ("String", "if {can_call}==false, output question for user to collecting enough information."),
                }
            }
        elif len(self.tool_dict.keys()) > 0:
            tool_results = await self.call_plan_func(self)
            if tool_results and len(tool_results.keys()) > 0:
                return {
                    "information": tool_results,
                    "instruction": ["Use format [keywords](url) to mark internet-source information in {OUTPUT}"]
                }
            else:
                return None
        else:
            return None

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": None,
            "alias": {
                "register_tool": { "func": self.register_tool },
                "call_tool": { "func": self.call_tool_func },
                "set_tool_proxy": { "func": self.set_tool_proxy },
                "add_public_tools": { "func": self.add_public_tools },
                "use_public_tools": { "func": self.add_public_tools },
                "add_public_categories": { "func": self.add_public_categories },
                "use_public_categories": { "func": self.add_public_categories },
                "add_all_public_tools": { "func": self.add_all_public_tools },
                "use_all_public_tools": { "func": self.add_all_public_tools },
                "must_call": { "func": self.must_call },
                "customize_call_plan": { "func": self.customize_call_plan },
            }
        }

def export():
    return ("Tool", Tool)


from .utils import FacilityABC
from Agently.utils import RuntimeCtx

class RoleManager(FacilityABC):
    def __init__(self, *, storage: object, plugin_manager: object, settings: object):
        self.storage = storage.table("role")
        self.plugin_manager = plugin_manager
        self.role_runtime_ctx = RuntimeCtx()

    def set_id(self, role_id: str):
        self.role_runtime_ctx.set("$ID", role_id)
        return self

    def set(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.set(key, value)
        else:
            self.role_runtime_ctx.set("DESC", key)
        return self

    def update(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.update(key, value)
        else:
            self.role_runtime_ctx.update("DESC", key)
        return self     

    def append(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.append(key, value)
        else:
            self.role_runtime_ctx.append("DESC", key)
        return self

    def extend(self, key: any, value: any=None):
        if value is not None:
            self.role_runtime_ctx.extend(key, value)
        else:
            self.role_runtime_ctx.extend("DESC", key)
        return self

    def save(self, role_id: str=None):
        role_id = role_id or self.role_runtime_ctx.get("$ID")
        if role_id is None or role_id == "":
            raise Exception("[Facility: RoleMananger] Role ID must be stated before save. Use .set_id() to specific it or pass role id into .save(<role_id>).")
        self.storage.set(role_id, self.role_runtime_ctx.get()).save()
        self.role_runtime_ctx.empty()
        return self

    def get(self, role_id: str):
        return self.storage.get(role_id)

def export():
    return ("role_manager", RoleManager)


import json
import inspect

class AliasManager(object):
    def __init__(self, target: object):
        self.target = target
        self.alias_name_list = []
        self.alias_details = {}

    def register(self, alias_name: str, alias_func: callable, *, return_value: bool=False, agent_component_name: str=None):
        if alias_name in self.alias_name_list:
            raise Exception(f"[Alias Manager] alias_name '{ alias_name }' has already been registered.")
        if return_value:
            setattr(self.target, alias_name, alias_func)
        else:
            def _alias_func(*args, **kwargs):
                alias_func(*args, **kwargs)
                return self.target
            setattr(self.target, alias_name, _alias_func)
        self.alias_name_list.append(alias_name)
        self.alias_details.update({ alias_name: { "func": alias_func, "return_value": return_value, "agent_component": agent_component_name } })
        return self

    def empty_alias(self):
        self.alias_name_list = []
        self.alias_details = {}
        return self

    def get_alias_info(self, *, group_by: str=None):
        result = {}
        if group_by == "agent_component":
            for alias_name, alias_info in self.alias_details.items():
                if alias_info["agent_component"] == None:
                    agent_component_name = "Unnamed"
                else:
                    agent_component_name = alias_info["agent_component"]
                if agent_component_name not in result:
                    result[agent_component_name] = {}
                result[agent_component_name][alias_name] = { "paramaters": [] }
                alias_func_info = inspect.signature(alias_info["func"])
                for name, meta in alias_func_info.parameters.items():
                    result[agent_component_name][alias_name]["paramaters"].append(str(meta.name))
                    result[agent_component_name][alias_name]["is_return_value"] = alias_info["return_value"]
                    result[agent_component_name][alias_name]["docstring"] = str(alias_info["func"].__doc__)
        else:
            for alias_name, alias_info in self.alias_details.items():
                result[alias_name] = { "paramaters": [] }
                alias_func_info = inspect.signature(alias_info["func"])
                for name, meta in alias_func_info.parameters.items():
                    result[alias_name]["paramaters"].append(str(meta.name))
                result[alias_name]["is_return_value"] = alias_info["return_value"]
                result[alias_name]["docstring"] = str(alias_info["func"].__doc__)
                result[alias_name]["agent_component"] = alias_info["agent_component"]
        return result

    def print_alias_info(self, *, group_by: str=None):
        print(json.dumps(self.get_alias_info(group_by=group_by), indent=4, ensure_ascii=False))


def find(data_list: list, find_func):
    """
    查找数据
    """
    found_data = None

    for data in data_list:
        if find_func(data):
            found_data = data
            break

    return found_data


def find_by_attr(data_list: list, attr_name: str, attr_val: any):
    """
    查找数据
    """
    found_data = None

    for data in data_list:
        if data.get(attr_name) == attr_val:
            found_data = data
            break

    return found_data


def find_all(data_list: list, find_func):
    """
    查找数据
    """
    found_data = []

    for data in data_list:
        if find_func(data):
            found_data.append(data)
            break

    return found_data


def find_all_by_attr(data_list: list, attr_name: str, attr_val: any):
    """
    查找数据
    """
    found_data = []

    for data in data_list:
        if data.get(attr_name) == attr_val:
            found_data.append(data)
            break

    return found_data

def has_target(data_list: list, judge_func):
    """
    判断数据是否存在
    """
    for data in data_list:
        if judge_func(data):
            return True

    return False


def has_target_by_attr(data_list: list, attr_name: str, attr_val: any):
    """
    判断数据是否存在
    """
    for data in data_list:
        if data.get(attr_name) == attr_val:
            return True

    return False



import json
import asyncio

from .utils import ComponentABC
from Agently.utils import to_json_desc, find_json

class Segment(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.is_enabled = lambda: self.agent.settings.get_trace_back("component_toggles.Segment")
        self.segments = {}
        self.response_segments_cache = {}
        self.parse_stage = 0
        self.current_segment = ""
        self.response_buffer = ""
        self.buffer_next_char = {
            "start": {
                "": "<",
                "<": "!",
                "<!": "%",
                "<!%": "%",
                "<!%%": "=",
            },
            "end": {
                "": "<",
                "<": "/",
                "</": "!",
                "</!": "%",
                "</!%": "%",
                "</!%%": ">",
            },
        }
        self.async_tasks = []  

    def add_segment(self, name: str, output_prompt: any, listener: callable=None, *, is_streaming = False, is_await = False):
        if name not in self.segments:
            self.segments[name] = {
                "prompt": output_prompt,
                "streaming_listeners": [],
                "done_listeners": [],
            }
            if listener != None:
                self.add_segment_listener(name, listener, is_streaming=is_streaming, is_await=is_await)
            return self.agent
        else:
            raise Exception(f"[Agent Component: Segment] segment name '{ name }' has already been used.")

    def add_segment_listener(self, name: str, listener: callable, *, is_streaming = False, is_await = False):
        if name in self.segments:
            if is_streaming:
                self.segments[name]["streaming_listeners"].append((listener, is_await))
            else:
                self.segments[name]["done_listeners"].append((listener, is_await))
        return self.agent

    def on_segment_delta(self, name: str, listener: callable, *, is_await = False):
        return self.add_segment_listener(name, listener, is_streaming = True, is_await= is_await)

    def on_segment_done(self, name: str, listener: callable, *, is_await = False):
        return self.add_segment_listener(name, listener, is_streaming = False, is_await= is_await)

    def _prefix(self):
        if not self.is_enabled() or len(self.segments) == 0:
            return None
        full_output_prompt = "- [OUTPUT RULE]: all output content of segments MUST START WITH TAG:<!%%={{segment_name}}>\n- [OUTPUT FORMAT]:\n"
        full_output_prompt += f"NOTICE: Segment Count={ len(self.segments.keys()) }, MUST MAKE SURE OUTPUT {{Segment Count}} PARTS OF SEGMENTS\n```\n"
        for name in self.segments:
            full_output_prompt += f"<!%%={ name }>\n{ to_json_desc(self.segments[name]['prompt']) }\n"
        full_output_prompt += "```\n(DO NOT OUTPUT```)\n"
        self.agent.output(full_output_prompt)

    async def _load_json_with_fix(self, origin: str):
        try:
            return json.loads(find_json(origin))
        except json.JSONDecodeError as e:
            try:
                fixed_result = await self.agent.request\
                    .input({
                        "origin JSON String": origin ,
                        "error": e.msg,
                        "position": e.pos,
                    })\
                    .output('Fixed JSON String can be parsed by Python only without explanation and decoration.')\
                    .get_result_async()
                fixed_result = json.loads(find_json(fixed_result))
            except Exception as e:
                raise Exception(f"[Agent Request] Error still occured when try to fix JSON decode error: { str(e) }")
        return fixed_result

    async def _suffix(self, event:str, data: any):
        if not self.is_enabled():
            return None
        try:
            if len(self.segments) > 0:
                self.segments = {}
                if event == "response:delta":
                    for char in data:
                        '''For Debug
                        print('char', char)
                        print('stage', self.parse_stage)
                        print('buffer', self.response_buffer)
                        print('cache', self.response_segments_cache)
                        print('--------------------')
                        '''
                        #Check Start
                        if self.parse_stage == 0:
                            if self.response_buffer in self.buffer_next_char["start"] and self.buffer_next_char["start"][self.response_buffer] == char:
                                self.response_buffer += char
                            if self.response_buffer == "<!%%=":
                                self.response_buffer = ""
                                self.parse_stage = 1
                                continue
                        #Record Segment Name
                        if self.parse_stage == 1:
                            if char != ">":
                                self.response_buffer += char
                            else:
                                self.current_segment = self.response_buffer
                                self.response_segments_cache[self.current_segment] = ""
                                self.response_buffer = ""
                                self.parse_stage = 2
                                continue
                        #Record Segment Content
                        if self.parse_stage == 2:
                            if self.response_buffer in self.buffer_next_char["start"] and self.buffer_next_char["start"][self.response_buffer] == char:
                                self.response_buffer += char
                            else:
                                self.response_segments_cache[self.current_segment] += self.response_buffer + char
                                self.response_buffer = ""
                                for (streaming_listener, is_await) in self.segments[self.current_segment]["streaming_listeners"]:
                                    if asyncio.iscoroutinefunction(streaming_listener):
                                        if is_await:
                                            await streaming_listener(char)
                                        else:
                                            self.async_tasks.append(asyncio.create_task(streaming_listener(char)))
                                    else:
                                        streaming_listener(char)
                            if self.response_buffer == "<!%%=":
                                segment_data = self.response_segments_cache[self.current_segment]
                                if isinstance(self.segments[self.current_segment]["prompt"], (dict, list, set)):
                                    segment_data = await self._load_json_with_fix(segment_data)
                                for (done_listener, is_await) in self.segments[self.current_segment]["done_listeners"]:
                                    if asyncio.iscoroutinefunction(done_listener):
                                        if is_await:
                                            await done_listener(done_segment_data)
                                        else:
                                            self.async_tasks.append(asyncio.create_task(done_listener(segment_data)))
                                    else:
                                        done_listener(segment_data)
                                self.response_buffer = ""
                                self.current_segment = ""
                                self.parse_stage = 1
                if event == "response:done":
                    segment_data = self.response_segments_cache[self.current_segment]
                    if isinstance(self.segments[self.current_segment]["prompt"], (dict, list, set)):
                        segment_data = await self._load_json_with_fix(segment_data)
                    for (done_listener, is_await) in self.segments[self.current_segment]["done_listeners"]:
                        if asyncio.iscoroutinefunction(done_listener):
                            if is_await:
                                await done_listener(done_segment_data)
                            else:
                                self.async_tasks.append(asyncio.create_task(done_listener(segment_data)))
                        else:
                            done_listener(segment_data)
                    self.agent.request.response_cache["reply"] = self.response_segments_cache
                if event == "response:finally":
                    for async_task in self.async_tasks:
                        await async_task
                    # clean request runtime
                    self.response_segments_cache = {}
                    self.parse_stage = 0
                    self.current_segment = ""
                    self.response_buffer = ""
        except Exception as e:
            self.segments = {}
            self.response_segments_cache = {}
            self.parse_stage = 0
            self.current_segment = ""
            self.response_buffer = ""
            raise(e)

    def export(self):
        return {
            "prefix": self._prefix,
            "suffix": self._suffix,
            "alias": {
                "segment": { "func": self.add_segment },
                "add_segment_listener": { "func": self.add_segment_listener },
                "on_segment_delta": { "func": self.on_segment_delta },
                "on_segment_done": { "func": self.on_segment_done },
            },
        }

def export():
    return ("Segment", Segment)


from .MainExecutor import MainExecutor
from .utils.exec_tree import resolve_runtime_data
from .Schema import Schema
from ..utils import RuntimeCtx
from .._global import global_settings
from .executors.install import mount_built_in_executors
from .lib.constants import EXECUTOR_TYPE_NORMAL

class Workflow:
    def __init__(self, *, schema_data: dict = None, settings: dict = {}):
        """
        Workflow，初始参数 schema_data 形如 { 'chunks': [], 'edges': [] }，handler 为要处理响应的函数
        """
        # 处理设置
        self.settings = RuntimeCtx(parent = global_settings)
        if settings:
            self.settings.update_by_dict(settings)
        # 初始 schema
        self.schema = Schema(schema_data or {'chunks': [], 'edges': []})
        # 初始化执行器
        self.executor = MainExecutor(settings)
        # 装载内置类型
        mount_built_in_executors(self.executor)
        # Chunk Storage
        self.chunks = {}

    def chunk(self, chunk_id: str, type=EXECUTOR_TYPE_NORMAL, **chunk_desc):
        def create_chunk_decorator(func: callable):
            return self.chunks.update({
                chunk_id: self.schema.create_chunk(
                        executor = func,
                        type = type,
                        **chunk_desc
                    )
            })
        return create_chunk_decorator
    
    def start(self):
        runtime_data = resolve_runtime_data(self.schema)
        self.executor.start(runtime_data)
    
    def reset(self, schema_data: dict):
        self.schema = Schema(schema_data or {'chunks': [], 'edges': []})